@@ -22,7 +22,10 @@ our @ISA = qw(Exporter);
 our @EXPORT = (@STATUS_CODES);
 our @EXPORT_OK = qw(%ERRORS);
 
-our $VERSION = $Nagios::Plugin::Functions::VERSION;
+# CPAN stupidly won't index this module without a literal $VERSION here,
+#   so we're forced to duplicate it explicitly
+our $VERSION = 0.17;
+# our $VERSION = $Nagios::Plugin::Functions::VERSION;
 
 sub new {
 	my $class = shift;
@@ -473,7 +476,8 @@ form "SHORTNAME CODE - $message".
 =item nagios_die( $message, [<CODE>] )
 
 Same as nagios_exit(), except that CODE is optional, defaulting
-to UNKNOWN.
+to UNKNOWN.  NOTE: exceptions are not raised by default to calling code.
+Set C<$_use_die> flag if this functionality is required (see test code).
 
 =item die( $message, [<CODE>] )
 
